from __future__ import annotations

from typing import Any, Callable, Concatenate, NoReturn, ParamSpec, TypeAlias, TypeVar

from pyferret import abstract, maybe

T = TypeVar("T", covariant=True)
E = TypeVar("E", covariant=True)
S = TypeVar("S")
U = TypeVar("U")
V = TypeVar("V", bound=object)
K = TypeVar("K", bound=object)
EXC = TypeVar("EXC", bound=Exception)
P = ParamSpec("P")


class Ok(abstract.Monad[T]):
    def fmap(self, func: Callable[[T], S]) -> Ok[S]:
        """
        If `Ok[T]` - applies `(T -> S)` and returns `Ok[S]`
        """
        return Ok(func(self._value))

    def fmap_partial(
        self,
        func: Callable[Concatenate[T, P], S],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Ok[S]:
        """
        If `Ok[T]` - applies `partial((T -> S), *args, **kwargs)` and returns `Ok[S]`
        """
        return Ok(func(self._value, *args, **kwargs))

    def fmap_through(self, func: Callable[[T], Any]) -> Ok[T]:
        """
        If `Ok[T]` - applies `(T -> Any)` to `T` and returns `Ok[T]`
        """
        _ = func(self._value)

        return self

    def fmap_partial_through(
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Ok[T]:
        """
        If `Ok[T]` - applies `partial((T -> Any), *args, **kwargs)` to `T` and returns
        `Ok[T]`
        """
        _ = func(self._value, *args, **kwargs)

        return self

    def bind(self, func: Callable[[T], Result[S, E]]) -> Result[S, E]:
        """
        If `Ok[T]` - applying `(T -> Result[S, E])`, and returns `Result[S, E]`
        """
        return func(self._value)

    def bind_partial(
        self,
        func: Callable[Concatenate[T, P], Result[S, E]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Result[S, E]:
        """
        If `Ok[T]` - applies `partial((T -> Result[S, E]), *args, **kwargs)` to `T` and
        returns `Result[S, E]`
        """
        return func(self._value, *args, **kwargs)

    def bind_through(self, func: Callable[[T], Result[S, E]]) -> Result[T, E]:
        """
        If `Ok[T]` - apply `(T -> Result[S, E])` and:
            - return `Ok[T]` if func result `Ok[Any]`
            - return `Err[E]` if func result `Err[E]`
        """
        result = func(self._value)

        if isinstance(result, Err):
            return result
        else:
            return self

    def bind_partial_through(
        self,
        func: Callable[Concatenate[T, P], Result[S, E]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Result[T, E]:
        """
        If `Ok[T]` - apply `partial((T -> Result[S, E]), *args, **kwargs)` and:
            - return `Ok[T]` if func result `Just[Any]`
            - return `Err[E]` if func result `Err[E]`
        """
        result = func(self._value, *args, **kwargs)

        if isinstance(result, Err):
            return result
        else:
            return self

    def bind_maybe(
        self: Ok[maybe.Maybe[S]], func: Callable[[S], Result[U, E]]
    ) -> Result[maybe.Maybe[U], E]:
        """
        If `Ok[Maybe[S]]` - apply `(S -> Result[U, E])` and:
            - return `Ok[Maybe[U]]` if func result `Ok[U]`
            - return `Err[E]` if func result `Err[E]`
        """
        if isinstance(self._value, maybe.Just):
            result = func(self._value._value)

            if isinstance(result, Ok):
                return Ok(maybe.Just(result._value))
            else:
                return result

        else:
            return Ok(maybe.Nothing())

    @property
    def is_err(self) -> bool:
        """
        If `Ok` return `False`
        """
        return False

    @property
    def is_ok(self) -> bool:
        """
        If `Ok` return `True`
        """
        return True

    @property
    def ok_value(self) -> T:
        """
        Unsafe return ok inner value
        """
        return self._value

    @property
    def err_value(self) -> NoReturn:
        """
        Raises ValueError
        """
        raise ValueError(f"Attempt to get err value on Ok: {self._value}")

    def get_ok_or(self, default: object) -> T:
        """
        Returns ok value
        """
        return self._value

    def get_err_or(self, default: S) -> S:
        """
        Returns default
        """
        return default

    def __repr__(self) -> str:
        return f"Ok {repr(self._value)}"


class Err(abstract.Monad[E]):
    def fmap(self, func: Callable[[V], K]) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def fmap_partial(
        self,
        func: Callable[Concatenate[V, P], K],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def fmap_through(self, func: Callable[[V], Any]) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def fmap_partial_through(
        self,
        func: Callable[Concatenate[V, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def bind(self, func: Callable[[V], Result[S, U]]) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def bind_partial(
        self,
        func: Callable[Concatenate[V, P], Result[S, E]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def bind_through(self, func: Callable[[V], Result[S, U]]) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def bind_partial_through(
        self,
        func: Callable[Concatenate[V, P], Result[S, E]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    def bind_maybe(self, func: Callable[[V], Result[S, U]]) -> Err[E]:
        """
        If `Err[E]` returns `Err[E]`
        """
        return self

    @property
    def is_err(self):
        """
        If `Err` return `True`
        """
        return True

    @property
    def is_ok(self):
        """
        If `Err` return `False`
        """
        return False

    @property
    def ok_value(self) -> NoReturn:
        """
        Raises ValueError
        Supports Exceptions and shows a cause if inner value is BaseException subclass
        """
        if isinstance(self._value, BaseException):
            raise ValueError(
                f"Attempt to get ok value on Err: {self._value.__class__.__name__}"
            ) from self._value

        raise ValueError(f"Attempt to get ok value on Err: {self._value}")

    @property
    def err_value(self) -> E:
        """
        Unsafe return err inner value
        """
        return self._value

    def get_ok_or(self, default: S) -> S:
        """
        Returns default
        """
        return default

    def get_err_or(self, default: object) -> E:
        """
        Returns err value
        """
        return self._value

    def __repr__(self) -> str:
        return f"Err {repr(self._value)}"


Result: TypeAlias = Ok[T] | Err[E]



from typing import Iterable, TypeVar

from pyferret import maybe

S = TypeVar("S")


def from_optional(value: S | None) -> maybe.Maybe[S]:
    if value is not None:
        return maybe.Just(value)
    else:
        return maybe.Nothing()


def concat(iterable: Iterable[Iterable[S]]) -> list[S]:
    return [item for sublist in iterable for item in sublist]



from .abstract import Applicative, Context, Functor, Monad
from .helpers import from_optional
from .maybe import Just, Maybe, Nothing
from .result import Err, Ok, Result

__all__ = [
    "Just",
    "Nothing",
    "Maybe",
    "Ok",
    "Err",
    "Result",
    "Context",
    "Functor",
    "Applicative",
    "Monad",
    "from_optional",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar

T = TypeVar("T", covariant=True)
S = TypeVar("S", covariant=True)


class Context(Generic[T]):
    """
    Base class of container that stores some value
    """

    __slots__ = ("_value",)

    def __init__(self, v: T) -> None:
        self._value = v

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and other._value == self._value

    def __ne__(self, __value: Any) -> bool:
        return not (self == __value)

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self._value))

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    def __rich_repr__(self):
        yield self._value


class Functor(Context[T]):
    @abstractmethod
    def fmap(self, func):
        """
        Applying `func` on inner value of context
        """

        raise NotImplementedError


class Applicative(Functor[T]):
    pass


class Monad(Applicative[T]):
    @abstractmethod
    def bind(self, func):
        """
        Applying `func` which returns a Monad on inner value of context and return its
        result
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Callable, Concatenate, NoReturn, ParamSpec, TypeAlias, TypeVar

from pyferret import abstract, result

T = TypeVar("T", covariant=True)
S = TypeVar("S")
U = TypeVar("U")
E = TypeVar("E")
V = TypeVar("V", bound=object)
K = TypeVar("K", bound=object)
P = ParamSpec("P")


class Just(abstract.Monad[T]):
    def fmap(self, func: Callable[[T], S]) -> Just[S]:
        """
        If `Just[T]` - applies `(T -> S)` to `T` and returns `Just[S]`
        """
        return Just(func(self._value))

    def fmap_through(self, func: Callable[[T], Any]) -> Just[T]:
        """
        If `Just[T]` - applies `(T -> Any)` to `T` and returns `Just[T]`
        """
        _ = func(self._value)
        return self

    def fmap_partial(
        self,
        func: Callable[Concatenate[T, P], S],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Just[S]:
        """
        If `Just[T]` - applies `partial((T -> S), *args, **kwargs)` to `T` and returns
        `Just[S]`
        """
        return Just(func(self._value, *args, **kwargs))

    def fmap_partial_through(
        self,
        func: Callable[Concatenate[T, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Just[T]:
        """
        If `Just[T]` - applies `partial((T -> Any), *args, **kwargs)` to `T` and returns
        `Just[T]`
        """
        _ = func(self._value, *args, **kwargs)

        return self

    def bind(self, func: Callable[[T], Maybe[S]]) -> Maybe[S]:
        """
        If `Just[T]` - applying `(T -> Maybe[S])`, and returns `Maybe[S]`
        """
        return func(self._value)

    def bind_through(self, func: Callable[[T], Maybe[S]]) -> Maybe[T]:
        """
        If `Just[T]` - apply `(T -> Maybe[S])` and:
            - return `Just[T]` if func result `Just[Any]`
            - return `Nothing` if func result `Nothing`
        """
        _ = func(self._value)
        return self

    def bind_partial(
        self,
        func: Callable[Concatenate[T, P], Maybe[S]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Maybe[S]:
        """
        If `Just[T]` - applies `partial((T -> Maybe[S]), *args, **kwargs)` to `T` and
        returns `Maybe[S]`
        """
        return func(self._value, *args, **kwargs)

    def bind_partial_through(
        self,
        func: Callable[Concatenate[T, P], Maybe[Any]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Maybe[T]:
        """
        If `Just[T]` - apply `partial((T -> Maybe[S]), *args, **kwargs)` and:
            - return `Just[T]` if func result `Just[Any]`
            - return `Nothing` if func result `Nothing`
        """
        _ = func(self._value, *args, **kwargs)
        return self

    def bind_result(
        self, func: Callable[[T], result.Result[S, E]]
    ) -> result.Result[Maybe[S], E]:
        """
        If `Just[T]` - apply (T -> Result[S, E]) and return Result[Maybe[S], E]
        """
        res = func(self._value)

        if isinstance(res, result.Ok):
            return result.Ok(Just(res._value))
        else:
            return res

    @property
    def is_some(self) -> bool:
        """
        Return `True` in case of `Just`
        """
        return True

    @property
    def value(self) -> T:
        """
        Unsafe return inner value in case of `Just`
        """
        return self._value

    def get_value_or(self, default: object) -> T:
        """
        If `Just` return inner value
        """
        return self._value

    def __repr__(self) -> str:
        return f"Just {repr(self._value)}"


class Nothing(abstract.Monad[None]):
    def __init__(self) -> None:
        self._value = None

    def fmap(self, func: Callable[[V], K]) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def fmap_partial(
        self,
        func: Callable[Concatenate[V, P], K],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def fmap_through(self, func: Callable[[V], Any]) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def fmap_partial_through(
        self,
        func: Callable[Concatenate[V, P], Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def bind(self, func: Callable[[V], Maybe[S]]) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def bind_partial(
        self,
        func: Callable[Concatenate[V, P], Maybe[S]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def bind_through(self, func: Callable[[V], Maybe[Any]]) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def bind_partial_through(
        self,
        func: Callable[Concatenate[V, P], Maybe[Any]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Nothing:
        """
        If `Nothing` returns `Nothing`
        """
        return self

    def bind_result(
        self, func: Callable[[V], result.Result[S, E]]
    ) -> result.Ok[Nothing]:
        """
        If `Nothing` returns `Ok[Nothing]`
        """
        return result.Ok(self)

    @property
    def is_some(self) -> bool:
        """
        Return `False` in case of `Nothing`
        """
        return False

    @property
    def value(self) -> NoReturn:
        """
        If `Nothing` raises an error
        """
        raise ValueError("Attempt to get value on Nothing")

    def get_value_or(self, default: S) -> S:
        """
        If `Just` return inner value
        """
        return default

    def __repr__(self) -> str:
        return "Nothing"


Maybe: TypeAlias = Just[T] | Nothing
